/*
 * WScout - Lightweight PCAP visualizer.
 * Copyright (C) 2008  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */

#include <cstring>
#include <sstream>
#include <iomanip>

extern "C"
{
#include <arpa/inet.h>
}

#include "wscout_gui_llc_header.hh"

namespace wscout
{

  namespace gui
  {

    bool
    LLCHeader::i_control::pf() const
    {
      return poll_final;
    }

    unsigned
    LLCHeader::i_control::snd_sn() const
    {
      return send_seqnum;
    }

    unsigned
    LLCHeader::i_control::rcv_sn() const
    {
      return receive_seqnum;
    }

    LLCHeader::s_control::supervisory_function
    LLCHeader::s_control::sf() const
    {
      return supervisory_function (superv_func);
    }

    bool
    LLCHeader::s_control::pf() const
    {
      return poll_final;
    }

    unsigned
    LLCHeader::s_control::rcv_sn() const
    {
      return receive_seqnum;
    }

    LLCHeader::u_control::command
    LLCHeader::u_control::cmd() const
    {
      return command (m2 << 4 | m1);
    }

    bool
    LLCHeader::u_control::pf() const
    {
      return poll_final;
    }

    LLCHeader::frame_format
    LLCHeader::format(const void* f, size_t caplen)
    {
      if (caplen < 3)
	return unknown;

      switch (static_cast<const uint8_t*> (f)[2] & 3)
	{
	case 0:
	case 2:
	  return information_transfer;
	case 1:
	  return supervisory;
	case 3:
	  return unnumbered;
	default:
	  return unknown;
	}
    }

    unsigned
    LLCHeader::as_byte(const dsap_addr& a)
    {
      return *reinterpret_cast<const uint8_t*> (&a);
    }

    unsigned
    LLCHeader::as_byte(const ssap_addr& a)
    {
      return *reinterpret_cast<const uint8_t*> (&a);
    }


    template <class C>
    void
    LLCHeader::setupDsapSsap(const frame<C>* f, size_t caplen)
    {
      if (caplen < sizeof (dsap_addr))
	{
	  dsapSsap_.setText(tr("truncated frame"));
	  return;
	}

      std::ostringstream os;

      os << std::hex << std::uppercase << "0x"
	 << std::setfill('0') << std::setw (2) << as_byte(f->dsap);
      if (f->dsap.ig == dsap_addr::group)
	os << " (group)";
      os << " < ";

      if (caplen < sizeof (dsap_addr) + sizeof (ssap_addr))
	{
	  os << "?? (truncated frame)";
	  dsapSsap_.setText(os.str().c_str());
	  return;
	}

      os << "0x" << std::setfill('0') << std::setw (2) << as_byte(f->ssap)
	 << (f->ssap.cr == ssap_addr::command ? " (command)" : " (response)");

      dsapSsap_.setText(os.str().c_str());
    }

    template <class C>
    LLCHeader::next_header
    LLCHeader::setup(const frame<C>*	f,
		     size_t		caplen,
		     std::ostream&	summary,
		     std::ostream&	control)
    {
      next_header	r (protocol::UNKNOWN, 0);

      if (f->ctl.pf())
	control << (f->ssap.cr == ssap_addr::command ? " (poll)" : " (final)");
      setupDsapSsap(f, caplen);
      if (f->dsap.addr == 0x55)
	{
	  const static uint8_t	snap_ip4[] = { 0x00, 0x00, 0x00, 0x08, 0x00 };
	  const static uint8_t	snap_ip6[] = { 0x00, 0x00, 0x00, 0x86, 0xDD };
	  const static uint8_t	snap_arp[] = { 0x00, 0x00, 0x00, 0x08, 0x06 };

	  summary << " + SNAP";

	  if (not memcmp(f->info, snap_ip4, sizeof (snap_ip4)))
	    {
	      summary << " (IPv4)";
	      r = next_header (protocol::IPv4, f->info + sizeof (snap_ip4));
	    }
	  else if (not memcmp(f->info, snap_ip6, sizeof (snap_ip6)))
	    {
	      summary << " (IPv6)";
	      r = next_header (protocol::IPv6, f->info + sizeof (snap_ip6));
	    }
	  else if (not memcmp(f->info, snap_arp, sizeof (snap_arp)))
	    {
	      summary << " (ARP)";
	      r = next_header (protocol::ARP, f->info + sizeof (snap_arp));
	    }
	}
      return r;
    }

    LLCHeader::next_header
    LLCHeader::setup(const void*		h,
		     size_t			caplen,
		     const Trace&		,
		     const frame_descriptor&	)
    {
      dsapSsap_.setText(tr("truncated frame"));

      next_header		r (protocol::UNKNOWN, 0);
      std::ostringstream	summary;
      std::ostringstream	control;

      summary << "LLC";
      switch (format(h, caplen))
	{
	case information_transfer:
	  {
	    const i_frame* const f = static_cast<const i_frame*> (h);

	    summary << " (I-PDU)";
	    control << f->ctl.snd_sn() << '-' << f->ctl.rcv_sn();
	    r = setup(f, caplen, summary, control);
	    break;
	  }

	case supervisory:
	  {
	    const s_frame* const f = static_cast<const s_frame*> (h);

	    summary << " (S-PDU)";
	    r = setup(f, caplen, summary, control);
	    break;
	  }

	case unnumbered:
	  {
	    const u_frame* const f = static_cast<const u_frame*> (h);

	    summary << " (U-PDU)";
	    switch (f->ctl.cmd())
	      {
	      case u_control::UI:	control << "UI";	break;
	      case u_control::DM:	control << "DM";	break;
	      case u_control::DISC:	control << "DISC";	break;
	      case u_control::UA:	control << "UA";	break;
	      case u_control::AC0:	control << "AC0";	break;
	      case u_control::FRMR:	control << "FRMR";	break;
	      case u_control::SABME:	control << "SABME";	break;
	      case u_control::XID:	control << "XID";	break;
	      case u_control::TEST:	control << "TEST";	break;
	      case u_control::AC1:	control << "AC1";	break;
	      default:			control << "INVALID";	break;
	      }
	    r = setup(f, caplen, summary, control);
	    break;
	  }

	default:
	  control << "??";
	  break;
	}
      summary_.setText(tr(summary.str().c_str()));
      control_.setText(tr(control.str().c_str()));
      return r;
    }

  } // End of namespace gui.

} // End of namespace wscout.
