/*
 * WScout - Lightweight PCAP visualizer.
 * Copyright (C) 2007, 2008  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */

#ifndef WSCOUT_GUI_TRACE_HXX_
# define WSCOUT_GUI_TRACE_HXX_

# include <QtGui/QScrollBar>
# include <boost/foreach.hpp>

# include "wscout_gui_trace.hh"

namespace wscout
{

  namespace gui
  {

    inline
    Trace::Trace(pcapxx::descriptor<>* trace, QWidget* parent):
      QAbstractScrollArea (parent),
      trace_ (0),
      swap_content_ (false),
      view_ (viewport()),
      last_selected_ (-1)
    {
      update_timer_.setSingleShot(true);
      connect(&update_timer_, SIGNAL(timeout()), SLOT(update()));
      setTrace(trace);
    }

    inline
    Trace::Trace(const Trace& t):
      QAbstractScrollArea (),
      trace_ (0),
      swap_content_ (false),
      view_ (viewport()),
      pos_ (t.pos_),
      selection_ (t.selection_),
      last_selected_ (t.last_selected_)
    {
      update_timer_.setSingleShot(true);
      connect(&update_timer_, SIGNAL(timeout()), SLOT(update()));
      if (const pcapxx::descriptor<>* p = t.trace())
	setTrace(new pcapxx::descriptor<> (*p));
    }

    inline
    void
    Trace::setTrace(pcapxx::descriptor<>* trace)
    {
      if (trace_ == trace)
	return;

      delete trace_;
      trace_ = trace;

      verticalScrollBar()->setRange(0, lastPosition());
      setPosition(0);
      emit traceChanged();
    }

    inline
    const pcapxx::descriptor<>*
    Trace::trace() const
    {
      return trace_;
    }

    inline
    pcapxx::descriptor<>*
    Trace::trace()
    {
      return trace_;
    }

    inline
    int
    Trace::position() const
    {
      return pos_;
    }

    inline
    int
    Trace::lastPosition() const
    {
      return trace_ ? trace_->size() - 1 : -1;
    }

    inline
    QString
    Trace::positionString(int p) const
    {
      if (not trace_)
	return tr("N/A");
      if (trace_->size() == 0)
	return tr("0 / 0 (100%)");
      if (p < 0)
	p = position();

      const int n = lastPosition();
      const int r = int (float (p + 1) / float (n + 1) * 100.);

      return tr("#%1 / %2 (%3%)").arg(p + 1).arg(n + 1).arg(r);
    }

    inline
    bool
    Trace::swapContent() const
    {
      return swap_content_;
    }

    inline
    int
    Trace::lastLeftClicked() const
    {
      return
	trace_ and
	0 <= last_left_clicked_ and last_left_clicked_ < int (trace_->size()) ?
	last_left_clicked_ : -1;
    }

    inline
    void
    Trace::startUpdateTimer(int ms)
    {
      if (not update_timer_.isActive())
	update_timer_.start(ms);
    }

    inline
    Trace::selection::range::range(int a, int b):
      std::pair<int, int> (std::min(a, b), std::max(a, b))
    {
    }

    inline
    void
    Trace::selection::clear()
    {
      ranges_.clear();
    }

    inline
    bool
    Trace::selection::contains(int i) const
    {
      BOOST_FOREACH(const range& r, ranges_)
	{
	  if (i <= r.second)
	    return i >= r.first;
	}
      return false;
    }

    inline
    std::ostream&
    Trace::selection::print(std::ostream& os) const
    {
      bool first = true;

      os << '{';
      BOOST_FOREACH(const range& r, ranges_)
	{
	  if (first)
	    first = false;
	  else
	    os << ", ";
	  os << r.first << '-' << r.second;
	}
      return os << '}';
    }

    inline
    Trace::selection&
    Trace::selection::toggle(int i)
    {
      int	lo = 0;
      iterator	it = ranges_.begin();

      while (it != ranges_.end() and i > it->second)
	{
	  lo = it->second;
	  ++it;
	}
      if (it == ranges_.end())
	{
	  if (ranges_.back().second == i - 1)
	    ++(ranges_.back().second);
	  else
	    ranges_.push_back(range (i, i));
	}
      else if (i == it->second)
	{
	  --(it->second);
	  cleanup(it);
	}
      else if (i < it->first - 1)
	ranges_.insert(it, range (i, i));
      else if (i == it->first - 1)
	--(it->first);
      else if (i == it->first)
	{
	  ++(it->first);
	  cleanup(it);
	}
      else
	{
	  ranges_.insert(it, range(it->first, i - 1));
	  it->first = i + 1;
	}
      return *this;
    }

    inline
    Trace::selection&
    Trace::selection::operator = (int i)
    {
      return *this = range (i, i);
    }

    inline
    Trace::selection&
    Trace::selection::operator = (const range& r)
    {
      clear();
      ranges_.push_back(r);
      return *this;
    }

    inline
    const Trace::selection::list_of_ranges&
    Trace::selection::ranges() const
    {
      return ranges_;
    }

    inline
    void
    Trace::selection::cleanup(iterator& it)
    {
      if (it->first > it->second)
	ranges_.erase(it);
    }

  } // End of namespace gui.

} // End of namespace wscout.

#endif // ! WSCOUT_GUI_TRACE_HXX_
