/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007, 2008  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */

#include <stdexcept>
#include <vector>
#include <cstdlib>

#include <boost/preprocessor/tuple/elem.hpp>
#include <boost/preprocessor/seq/for_each.hpp>

#include "config.hh"

#include <wipal/tool/options.hh>
#include <wipal/wifi/traces/similarity.hh>
#include <wipal/wifi/frame/filter/fast_intersector.hh>
#include <wipal/pcap/list_of_traces.hh>
#include <wipal/wifi/addr_mapping.hh>

namespace
{

  struct main_helper
  {
    template <class UniqueId>
    int
    operator () (const tool::opt::options& options) const
    {
      using wifi::frame::filter::fast_intersector;

      typedef pcapxx::list_of_traces<>	trace_type;
      typedef std::vector<trace_type>	trace_vector_type;
      typedef boost::tuple<unsigned, trace_vector_type::iterator,
				     trace_vector_type::iterator>
					score_type;
      typedef std::set<score_type>	score_set_type;

      wifi::addr_mapping	mapping (options.mapping);
      trace_vector_type		traces;
      score_set_type		scores;

      BOOST_FOREACH(trace_type t, options.input)
	{
	  if (options.prism_noise_means_error)
	    t.expect(pcapxx::descriptor<>::PRISM_HEADER);
	  traces.push_back(t);
	}

      if (traces.empty())
	return EXIT_SUCCESS;
      if (traces.front().empty())
	throw std::runtime_error ("First trace is empty");

      switch (traces.front().get(0).linktype())
	{

#define WP_case(R, Unused, LT)						\
	case pcapxx::descriptor<>::BOOST_PP_TUPLE_ELEM(2, 0, LT):	\
	  BOOST_FOREACH(const trace_type& t, traces)			\
	    t.expect(pcapxx::descriptor<>::BOOST_PP_TUPLE_ELEM(2, 0, LT)); \
	  scores = wifi::traces::similarity<UniqueId,			\
					    BOOST_PP_TUPLE_ELEM(2, 1, LT), \
					    fast_intersector>		\
	    (traces, mapping,						\
	     options.prism_noise_means_error, options.endianness);	\
	  break;

	  BOOST_PP_SEQ_FOR_EACH(WP_case, ~, WP_LINKTYPES)

#undef WP_case

	default:
	    tool::opt::unsupported_linktype(traces.front().get(0).linktype());
	}

      BOOST_FOREACH(const score_type& s, scores)
	std::cout << s.get<0>() << '\t'
		  << s.get<1>()->file_name() << '\t'
		  << s.get<2>()->file_name() << '\n';
      return EXIT_SUCCESS;
    }

  }; // End of struct main_helper.

} // End of anonymous namespace.

int
main(int argc, char *const argv[])
{
  try
    {
      const tool::opt::options	options =
	tool::opt::parse_args(argc, argv,
			      WP_OPT_ATTR
			      WP_OPT_MAPPING
			      WP_OPT_ENDIANNESS
			      WP_OPT_PRISM_NOISE
			      WP_OPT_USEC,
			      tool::opt::options::arbitrary_number_of_inputs,
			      0);
      return tool::opt::dispatch_unique_id(options, main_helper ());
    }
  catch (const std::exception& e)
    {
      std::cerr << "ERROR: " << e.what() << '.' << std::endl;
      return EXIT_FAILURE;
    }
}
