/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007, 2008  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */

#include <stdexcept>
#include <iostream>
#include <cstdlib>

#include <boost/preprocessor/tuple/elem.hpp>
#include <boost/preprocessor/seq/for_each.hpp>

#include <gmpxx.h>

#include "config.hh"

#include <wipal/tool/options.hh>
#include <wipal/pcap/dumper.hh>
#include <wipal/wifi/frame/filter/time_adjuster.hh>
#include <wipal/wifi/frame/filter/fast_intersector.hh>
#include <wipal/pcap/list_of_traces.hh>
#include <wipal/wifi/addr_mapping.hh>

namespace
{

  template <class T>
  struct sync_frames_dumper
  {
    sync_frames_dumper(pcapxx::dumper& dump): dump_ (dump)
    {
    }

    template <class TimeAdjuster>
    void	operator () (const TimeAdjuster& a)
    {
      a.for_each(dump_);
    }

  private:
    pcapxx::dumper&	dump_;
  };

  struct main_helper
  {
    template <class UniqueId>
    int
    operator () (const tool::opt::options& options) const
    {
      using wifi::frame::filter::provide_time_adjuster;

      mpf_set_default_prec(WP_GMP_FLOAT_PRECISION);

      wifi::addr_mapping	mapping (options.mapping);
      pcapxx::list_of_traces<>	traces1 (options.input[0]);
      pcapxx::list_of_traces<>	traces2 (options.input[1]);

      if (options.prism_noise_means_error)
	{
	  traces1.expect(pcapxx::descriptor<>::PRISM_HEADER);
	  traces2.expect(pcapxx::descriptor<>::PRISM_HEADER);
	}

      if (traces1.empty() or traces2.empty())
	throw std::invalid_argument ("Empty trace.");

      pcapxx::dumper		pcdump (traces2.get(0)
					.dumper(options.output[0]));

      sync_frames_dumper< pcapxx::list_of_traces<> >	dump (pcdump);

      switch (traces1.get(0).linktype())
	{
#define WP_case(R, Unused, LT)						\
	case pcapxx::descriptor<>::BOOST_PP_TUPLE_ELEM(2, 0, LT):	\
	  traces1.expect(pcapxx::descriptor<>::BOOST_PP_TUPLE_ELEM(2, 0, LT)); \
	  traces2.expect(pcapxx::descriptor<>::BOOST_PP_TUPLE_ELEM(2, 0, LT)); \
	  provide_time_adjuster<UniqueId, BOOST_PP_TUPLE_ELEM(2, 1, LT), \
	    wifi::frame::filter::fast_intersector>			\
	    (traces1.begin(), traces1.end(),				\
	     traces2.begin(), traces2.end(),				\
	     mapping, dump,						\
	     options.prism_noise_means_error,				\
	     options.endianness,					\
	     options.ref_blacklist);					\
	  break;

	  BOOST_PP_SEQ_FOR_EACH(WP_case, ~, WP_LINKTYPES)

#undef WP_case

	default:
	  tool::opt::unsupported_linktype(traces1.get(0).linktype());
	}
      return EXIT_SUCCESS;
    }

  }; // End of struct main_helper.

} // End of anonymous namespace.

int
main(int argc, char *const argv[])
{
  try
    {
      const tool::opt::options	options =
	tool::opt::parse_args(argc, argv,
			      WP_OPT_ATTR
			      WP_OPT_MAPPING
			      WP_OPT_ENDIANNESS
			      WP_OPT_PRISM_NOISE,
			      2, 1);
      return tool::opt::dispatch_unique_id(options, main_helper ());
    }
  catch (const std::exception& e)
    {
      std::cerr << "ERROR: " << e.what() << '.' << std::endl;
      return EXIT_FAILURE;
    }
}
